/**
 * @file boost/mirror/stream/simple_struct.hpp
 * @brief A structured object representation iostream writer implementation
 *
 *  Copyright 2008-2011 Matus Chochlik. Distributed under the Boost
 *  Software License, Version 1.0. (See accompanying file
 *  LICENSE_1_0.txt or copy at http://www.boost.org/LICENSE_1_0.txt)
 */

#ifndef BOOST_MIRROR_STREAM_STRUCTURED_1102141318_HPP
#define BOOST_MIRROR_STREAM_STRUCTURED_1102141318_HPP

#include <boost/mirror/stream/base.hpp>
#include <boost/mirror/cntnr_kind_tags.hpp>
#include <type_traits>

BOOST_MIRROR_NAMESPACE_BEGIN
namespace stream {

/// Wrapper class allowing to write objects in a structured format to ostreams
/**
 *  @see to_xml
 *  @see to_json
 */
struct to_simple_struct : _target<to_simple_struct> { };

template <typename T>
class _writer<T, to_simple_struct>
{
private:
        // helper functor executed on all members
        struct write_x
        {
                std::ostream& out;
                const _wrap<T, to_simple_struct>& obj;

                inline write_x(
                        std::ostream& os,
                        const _wrap<T, to_simple_struct>& ob
                ): out(os)
                 , obj(ob)
                { }

                static bool is_escaped_char(char c)
                {
                        return        (c == '"') ||
                                (c == '\\') ||
                                (c == '\b') ||
                                (c == '\f') ||
                                (c == '\n') ||
                                (c == '\r') ||
                                (c == '\t');
                }

                static char translate_esc_char(char c)
                {
                        switch(c)
                        {
                                case '\b': return 'b';
                                case '\f': return 'f';
                                case '\n': return 'n';
                                case '\r': return 'r';
                                case '\t': return 't';
                                default:;
                        }
                        return c;
                }

                void wr_val(bool x)
                {
                        out << x ? "true" : "false";
                }

                void wr_val(char x)
                {
                        out << '"';
                        if(is_escaped_char(x))
                                out << '\\' << translate_esc_char(x);
                        else out << x;
                        out << '"';
                }

                void wr_val(const std::string& x)
                {
                        out << '"';
                        std::string::const_iterator i = x.begin(), e = x.end();
                        while(i != e)
                        {
                                if(is_escaped_char(*i))
                                        out << '\\' << translate_esc_char(*i);
                                else out << *i;
                                ++i;
                        }
                        out << '"';
                }

                template <typename X>
                void wr_val(
                        const X& x,
                        typename std::enable_if<
                                std::is_arithmetic<X>::value
                        >::type* = nullptr
                )
                {
                        out << x;
                }

                template <typename X>
                void wr_val(
                        const X& x,
                        typename std::enable_if<
                                !std::is_arithmetic<X>::value &&
                                !std::is_enum<X>::value
                        >::type* = nullptr
                )
                {
                        out << to_simple_struct::from(x, obj);
                }

                template <typename X>
                void wr_val(
                        X x,
                        typename std::enable_if<
                                std::is_enum<X>::value
                        >::type* = nullptr
                )
                {
                        out << BOOST_MIRRORED_ENUM(X)::name_by_value(x);
                }
        };

        struct write_mem : write_x
        {
                inline write_mem(
                        std::ostream& os,
                        const _wrap<T, to_simple_struct>& ob
                ): write_x(os, ob)
                { }

                template <typename IterInfo>
                void operator()(IterInfo)
                {
                        typedef typename IterInfo::type meta_member;
                        // write the separator if necessary
                        if(!IterInfo::is_first::value) this->out << ", ";
                        // write the value
                        wr_val(meta_member::get(this->obj.value));
                }
        };

        // helper functor executed on all elements
        struct write_elem : write_x
        {
                inline write_elem(
                        std::ostream& os,
                        const _wrap<T, to_simple_struct>& ob
                ): write_x(os, ob)
                { }

                template <typename Locator>
                void operator()(Locator loc, bool first)
                {
                        // write the separator if necessary
                        if(!first) this->out << ", ";
                        // write the value
                        wr_val(loc.get());
                }
        };

        // write a structure
        template <typename MetaT>
        void write(
                MetaT,
                std::ostream& out,
                const _wrap<T, to_simple_struct>& obj,
                non_container_tag&
        )
        {
                static_assert(
                        mp::empty<all_containers<MetaT> >::value,
                        "Non-container types should not have containers"
                );
                out << "{";
                mp::for_each_ii<all_member_variables<MetaT> >(write_mem(out, obj));
                out << "}";
        }

        // write a simple container
        template <typename MetaT>
        void write(
                MetaT,
                std::ostream& out,
                const _wrap<T, to_simple_struct>& obj,
                simple_container_tag&
        )
        {
                out << "[";
                typedef typename default_traversal<
                        mp::front<all_containers<MetaT> >
                >::type meta_trav;
                //
                write_elem do_write(out, obj);
                auto trav=meta_trav::start(obj.value);
                if(!trav.done())
                {
                        do_write(trav.front(), true);
                        trav.step_front();
                        while(!trav.done())
                        {
                                do_write(trav.front(), false);
                                trav.step_front();
                        }
                }
                out << "]";
        }

        // other containers are not supported
        template <typename MetaT>
        void write(
                MetaT,
                std::ostream& out,
                const _wrap<T, to_simple_struct>& obj,
                container_tag&
        );

public:
        std::ostream& operator()(
                std::ostream& out,
                const _wrap<T, to_simple_struct>& obj
        )
        {
                typedef BOOST_MIRRORED_CLASS(T) meta_T;
                typename container_kind<meta_T>::type kind;
                write(meta_T(), out, obj, kind);
                return out;
        }

        std::ostream& operator()(
                std::ostream& out,
                const _wrap<T, to_simple_struct>& obj,
                const name_proc& namer
        )
        {
                typedef BOOST_MIRRORED_CLASS(T) meta_T;
                typename container_kind<meta_T>::type kind;
                out << '"'; namer(out); out << '"' << " = ";
                write(meta_T(), out, obj, kind);
                return out;
        }
};

} // namespace stream
BOOST_MIRROR_NAMESPACE_END

#endif //include guard

